/**********************************************************************
** Copyright (C) 2000-2002 Trolltech AS.  All rights reserved.
**
** This file is part of the Qtopia Environment.
**
** This file may be distributed and/or modified under the terms of the
** GNU General Public License version 2 as published by the Free Software
** Foundation and appearing in the file LICENSE.GPL included in the
** packaging of this file.
**
** This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
** WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
**
** See http://www.trolltech.com/gpl/ for GPL licensing information.
**
** Contact info@trolltech.com if any conditions of this licensing are
** not clear to you.
**
**********************************************************************/

// WARNING: Do *NOT* define this yourself. The SL5xxx from SHARP does NOT
//      have this class.
#define QTOPIA_INTERNAL_FSLP

#include "fileselector.h"
#include "fileselector_p.h"
#include "global.h"
#include "resource.h"
#include "config.h"
#include "applnk.h"
#include "storage.h"
#include "qpemenubar.h"
#include "qcopchannel_qws.h"
#include "lnkproperties.h"
#include "applnk.h"
#include "qpeapplication.h"
#include "categorymenu.h"

#include <stdlib.h>

#include <qdir.h>
#include <qwidget.h>
#include <qpopupmenu.h>
#include <qtoolbutton.h>
#include <qpushbutton.h>
#include <qheader.h>
#include <qtooltip.h>


FileSelectorItem::FileSelectorItem( QListView *parent, const DocLnk &f )
    : QListViewItem( parent ), fl( f )
{
    setText( 0, f.name() );
    setPixmap( 0, f.pixmap() );
}

FileSelectorItem::~FileSelectorItem()
{
}

class FileSelectorViewPrivate
{
public:
    CategoryMenu *cm;
};

FileSelectorView::FileSelectorView( const QString &f, QWidget *parent, const char *name )
    : QListView( parent, name ), filter( f ), count( 0 )
{
    d = new FileSelectorViewPrivate();
    d->cm = 0;
    setAllColumnsShowFocus( TRUE );
    addColumn( tr( "Name" ) );
    header()->hide();

    fileManager = new FileManager;
    reread();
    QCopChannel *channel = new QCopChannel( "QPE/Card", this );
    connect( channel, SIGNAL(received(const QCString &, const QByteArray &)),
	     this, SLOT(cardMessage( const QCString &, const QByteArray &)) );
}

FileSelectorView::~FileSelectorView()
{
}

void FileSelectorView::reread()
{
    FileSelectorItem *item = (FileSelectorItem *)selectedItem();
    QString oldFile;
    if ( item )
	oldFile = item->file().file();
    clear();
    DocLnkSet files;
    Global::findDocuments(&files, filter);
    count = files.children().count();
    QListIterator<DocLnk> dit( files.children() );
    for ( ; dit.current(); ++dit ) {
	if (d->cm)
	    if (!d->cm->isSelected((**dit).categories()))
		continue;
        item = new FileSelectorItem( this, **dit );
	if ( item->file().file() == oldFile )
	    setCurrentItem( item );
    }
    if ( !selectedItem() )
	setCurrentItem( firstChild() );
}

void FileSelectorView::setCategoryFilter(CategoryMenu *cm)
{
    d->cm = cm;
    connect(cm, SIGNAL(categoryChange()), this, SLOT(categoryChanged()) );
}

void FileSelectorView::categoryChanged() { reread(); }

void FileSelectorView::cardMessage( const QCString &msg, const QByteArray &)
{
    if ( msg == "mtabChanged()" )
	reread();
}

void FileSelectorView::keyPressEvent( QKeyEvent *e )
{
    QString txt = e->text();
    if (e->key() == Key_Space)
	emit returnPressed( currentItem() );
    else if ( !txt.isNull() && txt[0] > ' ' && e->key() < 0x1000 )
	e->ignore();
    else
	QListView::keyPressEvent(e);
}

class FileSelectorPrivate
{
public:
    CategoryMenu *cm;
    QMenuBar *mb;
};

/*!
  \class FileSelector fileselector.h
  \brief The FileSelector widget allows the user to select DocLnk objects.

  This class presents a file selection dialog to the user. Use
  setNewVisible() depending on whether the application can be used to
  create new files or not. Use setCloseVisible() depending on
  whether the user may leave the dialog without creating or selecting
  a document or not. The number of files in the view is available from
  fileCount(). To force the view to be updated call reread().

  If the user presses the 'New Document' button the newSelected()
  signal is emitted. If the user selects an existing file the
  fileSelected() signal is emitted. The selected file's \link
  doclnk.html DocLnk\endlink is available from the selected()
  function. If the file selector is no longer necessary the closeMe()
  signal is emitted.
*/

/*!
  Constructs a FileSelector with mime filter \a f.
  The standard Qt \a parent and \a name parameters are passed to the
  parent widget.

  If \a newVisible is TRUE, the widget has a button to allow the user
  the create "new" documents; this is useful for applications that can
  create and edit documents but not suitable for applications that
  only provide viewing.

  If \a closeVisible is TRUE, the widget has a button to allow the
  user to close the dialog without creating or selecting a document,
  i.e. this effectively allows the user to select "no document".

  \sa DocLnkSet::DocLnkSet()
*/
FileSelector::FileSelector( const QString &f, QWidget *parent, const char *name, bool newVisible, bool closeVisible )
    : QVBox( parent, name ), filter( f )
{
    setMargin( 0 );
    setSpacing( 0 );
    QHBox *top = new QHBox( this );
    top->setBackgroundMode( PaletteButton );	// same colour as toolbars
    top->setSpacing( 0 );

    QWidget *spacer = new QWidget( top );
    spacer->setBackgroundMode( PaletteButton );

    d = new FileSelectorPrivate();
    d->mb = new QMenuBar(spacer);
    d->cm = new CategoryMenu("Document View", this);
    QPEMenuToolFocusManager::manager()->addWidget( d->mb );
    d->mb->insertItem(tr("View"), d->cm);


    QToolButton *tb = new QToolButton( top );
    tb->setPixmap( Resource::loadPixmap( "new" ) );
    connect( tb, SIGNAL( clicked() ), this, SLOT( createNew() ) );
    buttonNew = tb;
    tb->setFixedSize( 18, 20 ); // tb->sizeHint() );
    tb->setAutoRaise( TRUE );
    QToolTip::add( tb, tr( "Create a new Document" ) );
    QPEMenuToolFocusManager::manager()->addWidget( tb );

    tb = new QToolButton( top );
    tb->setPixmap( Resource::loadPixmap( "close" ) );
    connect( tb, SIGNAL( clicked() ), this, SIGNAL( closeMe() ) );
    buttonClose = tb;
    tb->setFixedSize( 18, 20 ); // tb->sizeHint() );
    tb->setAutoRaise( TRUE );
    QToolTip::add( tb, tr( "Close the File Selector" ) );
    QPEMenuToolFocusManager::manager()->addWidget( tb );

    view = new FileSelectorView( filter, this, "fileview" );
    view->setCategoryFilter(d->cm);
    QPEApplication::setStylusOperation( view->viewport(), QPEApplication::RightOnHold );
    connect( view, SIGNAL( mouseButtonClicked( int, QListViewItem *, const QPoint &, int ) ),
	     this, SLOT( fileClicked( int, QListViewItem *, const QPoint &, int ) ) );
    connect( view, SIGNAL( mouseButtonPressed( int, QListViewItem *, const QPoint &, int ) ),
	     this, SLOT( filePressed( int, QListViewItem *, const QPoint &, int ) ) );
    connect( view, SIGNAL( returnPressed( QListViewItem * ) ),
	     this, SLOT( fileClicked( QListViewItem * ) ) );

    setNewVisible( newVisible );
    setCloseVisible( closeVisible );
}

/*!
  Destroys the widget.
*/
FileSelector::~FileSelector()
{

}

/*!
  Returns the number of files in the view. If this is zero, an editor
  application might bypass the selector and immediately start with
  a "new" document.
*/
int FileSelector::fileCount()
{
    return view->fileCount();
}

/*!
  Calling this function is the programmatic equivalent of the user
  pressing the "new" button.

  \sa newSelected(), closeMe()
*/
void FileSelector::createNew()
{
    DocLnk f;
    emit newSelected( f );
    emit closeMe();
}

void FileSelector::fileClicked( int button, QListViewItem *i, const QPoint &, int )
{
    if ( !i )
	return;
    if ( button == Qt::LeftButton ) {
	fileClicked( i );
    }
}

void FileSelector::filePressed( int button, QListViewItem *i, const QPoint &, int )
{
    if ( !i )
	return;
    if ( button == Qt::RightButton ) {
	DocLnk l = ((FileSelectorItem *)i)->file();
	LnkProperties prop( &l );
	prop.showMaximized();
	prop.exec();
	d->cm->reload();
	reread();
    }
}

void FileSelector::fileClicked( QListViewItem *i )
{
    if ( !i )
	return;
    emit fileSelected( ( (FileSelectorItem*)i )->file() );
    emit closeMe();
}

/*!
  Returns the selected \link doclnk.html DocLnk\endlink. The caller is
  responsible for deleting the returned value.
*/
const DocLnk *FileSelector::selected()
{
    FileSelectorItem *item = (FileSelectorItem *)view->selectedItem();
    if ( item )
	return new DocLnk( item->file() );
    return NULL;
}

/*!
  \fn void FileSelector::fileSelected( const DocLnk &f )

  This signal is emitted when the user selects a document.
  \a f is the document.
*/

/*!
  \fn void FileSelector::newSelected( const DocLnk &f )

  This signal is emitted when the user selects a "new" document.
  \a f is a DocLnk for the document. You will need to set the type
  of the document after copying it.
*/

/*!
  \fn void FileSelector::closeMe()

  This signal is emitted when the user no longer needs to view the widget.
*/


/*!
  If \a b is TRUE a "new document" button is visible; if \a b is FALSE
  this button is not visible and the user is unable to create new
  documents from the dialog.
*/
void FileSelector::setNewVisible( bool b )
{
    if ( b )
	buttonNew->show();
    else
	buttonNew->hide();
}

/*!
  If \a b is TRUE a "close" or "no document" button is visible; if \a
  b is FALSE this button is not visible and the user is unable to
  leave the dialog without creating or selecting a document.
*/
void FileSelector::setCloseVisible( bool b )
{
    if ( b )
	buttonClose->show();
    else
	buttonClose->hide();
}

/*!
  Rereads the list of documents.
*/
void FileSelector::reread()
{
    view->reread();
}
